'use strict';

process.env.BABEL_ENV = 'development';

const webpack = require('webpack');
const chalk = require('chalk');
const clearConsole = require('react-dev-utils/clearConsole');
const formatWebpackMessages = require('react-dev-utils/formatWebpackMessages');
const createWdm = require('webpack-dev-middleware');
const c2k = require('koa2-connect');
const vm = require('vm');
const path = require('path');

exports.createExpressMiddleware = createExpressMiddleware;
function createExpressMiddleware(options, koaContextLocals) {
  const { root } = options || {};
  if (!root) {
    console.warn('byted-react-scripts middleware need options.root to be set. Please check your config.');
    return (req, res, next) => { next(); };
  }
  process.env.BRS_PROJECT_ROOT = root;

  let ssrBundleModule;
  let injectStatsRef = { current: undefined };

  // 必须晚于 env.BRS_PROJECT_ROOT 设置完成，子依赖执行顺序问题
  const getWebpackSsrConfig = require('../config/webpack.config.ssr');
  const getWebpackDevConfig = require('../config/webpack.config.dev');
  const compiler = webpack([
    getWebpackSsrConfig(root),
    getWebpackDevConfig({
      isWatch: true
    })]
  );

  const wdm = createWdm(
    compiler,
    { serverSideRender: true }
  );
  const fs = compiler.compilers[0].outputFileSystem;

  compiler.hooks.done.tap('done', stats => {
    clearConsole();
    injectStatsRef.current = stats;
    const messages = formatWebpackMessages(
      stats.toJson({ all: false, warnings: true, errors: true })
    );
    const isSuccessful = !messages.errors.length && !messages.warnings.length;
    if (isSuccessful) {
      console.log(chalk.green('Compiled successfully!'));
    }

    const filename = process.env.SSR_BUNDLE_FILENAME || path.resolve(root, 'build/static/bundle.ssr.js');
    const code =
      fs.readFileSync(filename).toString('utf8');
    const dummy = {};
    try {
      vm.runInNewContext(code, {
        exports: dummy,
        require,
        global,
        setTimeout,
        process,
        clearTimeout,
        setInterval,
        clearInterval,
        displayErrors: true
      });
    } catch (e) {
      console.error(e);
      process.exit(1);
    }
    ssrBundleModule = dummy;

    if (messages.errors.length) {
      if (messages.errors.length > 1) {
        messages.errors.length = 1;
      }
      console.log(chalk.red('Failed to compile.\n'));
      console.log(messages.errors.join('\n\n'));
      return;
    }

    if (messages.warnings.length) {
      console.log(chalk.yellow('Compiled with warnings.\n'));
      console.log(messages.warnings.join('\n\n'));

      console.log(
        '\nSearch for the ' +
        chalk.underline(chalk.yellow('keywords')) +
        ' to learn more about each warning.'
      );
      console.log(
        'To ignore, add ' +
        chalk.cyan('// eslint-disable-next-line') +
        ' to the line before.\n'
      );
    }
  });
  return (req, res, next) => {
    if (res.locals) {
      res.locals.ssrBundleModule = ssrBundleModule;
    } else if (koaContextLocals) {
      koaContextLocals.ssrBundleModule = ssrBundleModule;
      Object.assign(koaContextLocals, {
        fs,
        webpackStats: injectStatsRef.current,
        webpackCsrStats: injectStatsRef.current.toJson().assetsByChunkName
      });
    }
    wdm(req, res, next);
  }
}

exports.createKoaMiddleware = createKoaMiddleware;
function createKoaMiddleware(options) {
  let ctx = {};
  const middleware = c2k(createExpressMiddleware(options, ctx));
  return (koaContext, next) => {
    koaContext.locals = koaContext.locals || {};
    koaContext.locals.wdm = ctx;
    return middleware(koaContext, next);
  };
}

/*
// test code
const express = require('express');
const app = express();
app.use(createExpressMiddleware({
  root: process.cwd()
}));

app.get('/*', (req, res) => {
  res.send(res.locals.ssrBundleModule.renderAppToString());
})

app.listen(3001, () => {
  console.log('server started at port 3001.');
});
*/
