'use strict';
/*
 * 专门为 webpack 打包的项目生产 dts
 *
 * usage: node dts.js <rootDir>
 * @param {string} rootDir 项目根目录
 *
 * 通过项目根目录的 package.json 信息拿到 bundle 文件路径 (pkg.main)
 * 在 bundle 文件同目录下生成同名的 d.ts 文件
 */

const path = require('path');
const fs = require('fs-extra');
const ts = require('typescript');
const chalk = require('chalk');
const paths = require('../../config/paths');
const getTsConfigSync = require('../utils/readTsconfigSync');
const printBuildError = require('react-dev-utils/printBuildError');
const { sendLog } = require('../utils/log');
process.on('unhandledRejection', err => {
  printBuildError(err);
  process.exit(1);
});

module.exports = async (root, targetRoot) => {
  const start = process.hrtime();
  const compilerOptions = getTsConfigSync(paths.appTsConfig);
  const pkgFilename = path.resolve(root, 'package.json');
  const pkgConfig = require(pkgFilename);
  const ns = pkgConfig.name;
  let entryFilename = path.resolve(root, path.resolve(root, 'src/ssr'));
  if (fs.existsSync(entryFilename + '.ts')) {
    entryFilename += '.ts';
  } else if (fs.existsSync(entryFilename + '.tsx')) {
    entryFilename += '.tsx';
  } else {
    console.log(chalk.yellow('[BUILD dts]: ') + 'entry file not found.');
    return;
  }

  console.log(chalk.yellow('[BUILD dts]: ') + `entry: ${entryFilename}`);
  const bundleFilename = path.resolve(targetRoot, 'typings/ssr.d.ts');
  const program = ts.createProgram([entryFilename], {
    ...compilerOptions,
    emitDeclarationOnly: true,
    declaration: true,
    module: ts.ModuleKind.CommonJS,
    target: ts.ScriptTarget.ESNext,
    lib: ['lib.dom.d.ts', 'lib.esnext.full.d.ts'],
    suppressOutputPathCheck: true,
    moduleResolution: ts.ModuleResolutionKind.NodeJs,
    outFile: bundleFilename,
    rootDir: undefined  // 防止 monorepo 工程源码编译的问题
  });
  const emitResult = program.emit();
  const allDiagnostics = ts.getPreEmitDiagnostics(program).concat(emitResult.diagnostics);
  allDiagnostics.forEach(diagnostic => {
    if (diagnostic.file) {
      const { line, character } = diagnostic.file.getLineAndCharacterOfPosition(diagnostic.start);
      const message = ts.flattenDiagnosticMessageText(diagnostic.messageText, '\n');
      if (message.match(/Cannot find module.+/)) {
        return;
      }
      console.log(`${diagnostic.file.fileName} (${line + 1},${character + 1}): ${message}`);
    } else {
      console.log(`${ts.flattenDiagnosticMessageText(diagnostic.messageText, '\n')}`);
    }
  });
  if (!fs.existsSync(bundleFilename)) {
    console.log(chalk.bgYellowBright('generate bundle.ssr.d.ts failed.'));
    // throw new Error(`typings field not found in ${bundleFilename}.`);
    return;
  }
  const sourceText = fs.readFileSync(bundleFilename, 'utf8');
  let ast = ts.createSourceFile('', sourceText, ts.ScriptTarget.ESNext);
  const traverseState = {
    changes: {}
  };
  let entryNode;
  ast.statements.forEach(d => {
    // 获取相对路径
    // const relativeRoot = path.relative(compilerOptions.baseUrl, path.resolve(root, 'src'));
    if (ts.isModuleDeclaration(d)) {
      // 计算 relativeRoot 的 edge case 较多，还没想清楚，先粗暴一点吧...
      if (d.name.text === 'ssr' || d.name.text.indexOf('src/ssr') !== -1) {
        entryNode = d;
      }
      // TODO: 路径改写这块 monorepo 还没有好的办法，先要拿到所有包的正确路径映射关系
      // 不过目前不正确映射也没有关系，不影响主要功能
      const old = d.name.text;
      const n = `${ns}/${old}`.replace(/\.tsx?$/, '').replace(/\/index$/, ''); // 把 index 全部干掉
      traverseState.changes[old] = n;
    }
  });
  // let cnt: any = {};
  const prt = ts.createPrinter(
    {
      newLine: ts.NewLineKind.CarriageReturnLineFeed,
      noEmitHelpers: true,
      omitTrailingSemicolon: true,
      removeComments: false
    },
    {
      substituteNode(_hint, node) {
        // cnt[node.kind] = cnt[node.kind] || 0;
        // cnt[node.kind]++;
        if (
          // declare module 'xxxxx'
          ts.isModuleDeclaration(node) &&
          (node.flags & ts.NodeFlags.Namespace) !== ts.NodeFlags.Namespace // 过滤掉 declare namespace（ts 这里的 flag 居然没有提供判断用的 api）
        ) {
          const rtn = appendNamespaceToModuleName(node);
          return rtn;
        } else if (
          //  import {xxxx} from 'xxxxx'
          ts.isImportDeclaration(node) &&
          node.moduleSpecifier.text
        ) {
          return replaceImportModuleName(node);
        } else if (
          // export {xxxxx as xxxx} from 'xxxxx'
          ts.isExportDeclaration(node) &&
          node.moduleSpecifier &&
          ts.isStringLiteralLike(node.moduleSpecifier)
        ) {
          return replaceExportDeclarationNode(node);
        } else if (
          // var a: import('xxxxx').yyyy = xxxx;
          ts.isImportTypeNode(node)
        ) {
          return replaceImportTypeNode(node);
        } else if (
          // var a: T1<import('xxxx').yyyyy>> = xxxxx;
          ts.isLiteralTypeNode(node)
        ) {
          return replaceLiteralTypeNode(node);
        }
        return node;
      }
    }
  );
  // const namespaceToExport = ns.replace(/\//g, '_').replace(/@/g, '');
  const nsNode = ts.createModuleDeclaration(
    [],
    [ts.createToken(ts.SyntaxKind.DeclareKeyword)],
    ts.createIdentifier('build/static/bundle.ssr'),
    entryNode.body
  );
  ast = ts.updateSourceFileNode(
    ast,
    ast.statements.concat(
      nsNode
      // ts.createExportAssignment(
      //   [],
      //   [],
      //   true,
      //   ts.createIdentifier(namespaceToExport)
      // )
    )
  );
  const code = prt.printFile(ast);
  await fs.writeFile(bundleFilename, code, 'utf8');
  console.log('finished.');
  const [_s, _ns] = process.hrtime(start);
  const _t = _s * 1000 + _ns * 1e-6;
  sendLog(_t, 'ssr').catch();

  function appendNamespaceToModuleName(node) {
    const old = node.name.text;
    const n = `${ns}/${old}`.replace(/\.tsx?$/, '').replace(/\/index$/, ''); // 把 index 全部干掉
    traverseState.changes[old] = n;
    return ts.updateModuleDeclaration(node, node.decorators, node.modifiers, ts.createStringLiteral(n), node.body);
  }
  function replaceImportModuleName(node) {
    const old = node.moduleSpecifier.text;
    if (traverseState.changes[old]) {
      // console.log(`${old} => ${traverseState.changes[old]}`);
      return ts.updateImportDeclaration(
        node,
        node.decorators,
        node.modifiers,
        node.importClause,
        ts.createStringLiteral(traverseState.changes[old])
      );
    }
    return node;
  }
  function replaceLiteralTypeNode(node) {
    const newModuleName = traverseState.changes[node.literal.text];
    if (!newModuleName) {
      return node;
    }
    return ts.updateLiteralTypeNode(node, ts.createStringLiteral(newModuleName));
  }
  function replaceExportDeclarationNode(node) {
    const newModuleName = traverseState.changes[node.moduleSpecifier.text];
    if (!newModuleName) {
      return node;
    }
    return ts.updateExportDeclaration(
      node,
      node.decorators,
      node.modifiers,
      node.exportClause,
      ts.createStringLiteral(newModuleName)
    );
  }
  function replaceImportTypeNode(node) {
    const newModuleName = traverseState.changes[node.argument.literal.text];
    if (!newModuleName) {
      return node;
    }
    return ts.updateImportTypeNode(
      node,
      ts.createStringLiteral(newModuleName),
      node.qualifier,
      node.typeArguments,
      node.isTypeOf
    );
  }
};
