'use strict';

require('../../config/env');
const printHostingInstructions = require('react-dev-utils/printHostingInstructions');
const formatWebpackMessages = require('react-dev-utils/formatWebpackMessages');
const webpack = require('webpack');
const paths = require('../../config/paths');
const chalk = require('chalk');
const FileSizeReporter = require('react-dev-utils/FileSizeReporter');
const fs = require('fs');
const useYarn = fs.existsSync(paths.yarnLockFile);
const commander = require('../utils/commander');
const { mergeWebpackConfig } = require('../../config/mergeWebpackConfig');
const { sendLog } = require('../utils/log');

// These sizes are pretty large. We'll warn for bundles exceeding them.
const WARN_AFTER_BUNDLE_GZIP_SIZE = 512 * 1024;
const WARN_AFTER_CHUNK_GZIP_SIZE = 1024 * 1024;
const outDir = process.argv.slice(2)[0] || paths.appBuild;
const webpackOptions = {
  isProduction: true,
  useHash: !commander.no_hash,
  useModern: !!commander.modern,
  useSnap: !!commander.snap
}
const createWebpackProdConfig = require('../../config/webpack.config.library');
const webpackConfig = mergeWebpackConfig(createWebpackProdConfig(webpackOptions));
webpackConfig.output.path = outDir;
const compiler = webpack(webpackConfig);

log('START...');
const start = process.hrtime();
compiler.run((err, stats) => {
  const [s, ns] = process.hrtime(start);
  if (err) {
    console.log(err);
  }
  const _t = s * 1000 + ns * 1e-6;
  sendLog(_t, 'lib').catch();
  log('END cost ' + chalk.yellow(s + 'sec ' + (ns * 1e-6).toFixed(3) + 'msec'));
  let messages;
  if (err) {
    messages = formatWebpackMessages({
      errors: [err.message],
      warnings: []
    });
  } else {
    messages = formatWebpackMessages(
      stats.toJson({ all: false, warnings: true, errors: true })
    );
  }

  if (messages.errors.length) {
    // Only keep the first error. Others are often indicative
    // of the same problem, but confuse the reader with noise.
    if (messages.errors.length > 1) {
      messages.errors.length = 1;
    }
    return;
  }
  reportBuildResult({
    stats,
    warnings: messages.warnings,
    previousFileSizes: JSON.parse(process.env.previousFileSizes) || {}
  });
});

function reportBuildResult(reports) {
  const { stats, previousFileSizes, warnings } = reports;
  if (warnings.length) {
    log(chalk.yellow('Compiled with warnings.\n'));
    log(warnings.join('\n\n'));
    log(
      '\nSearch for the ' +
      chalk.underline(chalk.yellow('keywords')) +
      ' to learn more about each warning.'
    );
    log(
      'To ignore, add ' +
      chalk.cyan('// eslint-disable-next-line') +
      ' to the line before.\n'
    );
  } else {
    log(chalk.green('Compiled successfully.\n'));
  }

  log('File sizes after gzip:\n');

  FileSizeReporter.printFileSizesAfterBuild(
    stats,
    previousFileSizes,
    paths.appBuild,
    0,
    WARN_AFTER_BUNDLE_GZIP_SIZE,
    WARN_AFTER_CHUNK_GZIP_SIZE
  );

  const appPackage = require(paths.appPackageJson);
  const publicUrl = paths.publicUrl || '';
  printHostingInstructions(appPackage, publicUrl, '', paths.appBuild, useYarn);

  // if (webpackOptions.useSnap) {
  //   console.log(`${chalk.cyan('Run snapshot...')}`);
  //   require('react-snap/run');
  // }
}

function log(str) {
  console.log(`${chalk.cyanBright('[BUILD Library]: ')}${str}`);
}
