'use strict';

const cwd = process.env.BRS_PROJECT_ROOT || process.cwd();
const fs = require('fs-extra');
const path = require('path');
const chalk = require('chalk');
const paths = require('../../config/paths');
const { mergeWebpackConfig } = require('../../config/mergeWebpackConfig');
const { sendLog } = require('../utils/log');

// 检查 SSR 入口文件是否存在
if (
  !fs.existsSync(path.resolve(paths.appSrc, 'ssr.js')) &&
  !fs.existsSync(path.resolve(paths.appSrc, 'ssr.ts')) &&
  !fs.existsSync(path.resolve(paths.appSrc, 'ssr.jsx')) &&
  !fs.existsSync(path.resolve(paths.appSrc, 'ssr.tsx'))
) {
  console.warn(chalk.yellow('SSR entry file (src/ssr.(t|j)sx) not found.'));
  process.exit(0);
}
const webpack = require('webpack');
const getSsrWebpackConfig = require('../../config/webpack.config.ssr');
const buildDefinitionFileForSSR = require('./buildDefinitionFileForSSR');
const ssrWebpackConfig = mergeWebpackConfig(getSsrWebpackConfig(cwd));

const outDir = process.argv.slice(2)[0] || paths.appBuild;

log(`START...`);
const start = process.hrtime();
ssrWebpackConfig.output.path = outDir;
const compiler = webpack(ssrWebpackConfig);
compiler.run((err, stats) => {
  const [s, ns] = process.hrtime(start);
  if (err) {
    console.log(err);
  }
  const _t = s * 1000 + ns * 1e-6;
  sendLog(_t, 'ssr').catch();
  log(`END cost ` + chalk.yellow(`${s}sec ${(ns * 1e-6).toFixed(3)}msec`));
  if (err || stats.hasErrors()) {
    stats.compilation.errors.forEach(e => {
      log(e.message);
    });
    log(chalk.red('ssr build failed.'));
    process.exit(1);
  } else {
    log(chalk.green('Asserts for server render compiled successfully:'));
    Object.keys(stats.compilation.assets).forEach(name => {
      if (/\.js$/.test(name)) {
        log('    - ' + chalk.whiteBright(name));
      }
    });
  }
  try {
    require(`${outDir}/static/bundle.ssr`);
  } catch (e) {
    log('SSR bundle has Error.');
    console.error(e);
    process.exit(1);
  }
  if (typescriptInstalled()) {
    log('Start building definistion file [typings/ssr.d.ts]...');
    buildDefinitionFileForSSR(paths.appPath, outDir);
    return;
  }

  process.exit(0);
});

function log(str) {
  console.log(`${chalk.yellowBright('[BUILD SSR]: ')}${str}`);
}

function typescriptInstalled() {
  try {
    require.resolve('typescript');
    return true;
  } catch (e) {
    return false;
  }
}
