// @remove-on-eject-begin
/**
 * Copyright (c) 2015-present, Facebook, Inc.
 *
 * This source code is licensed under the MIT license found in the
 * LICENSE file in the root directory of this source tree.
 */
// @remove-on-eject-end
'use strict';

// Do this as the first thing so that any code reading it knows the right env.
process.env.BABEL_ENV = 'production';
process.env.NODE_ENV = 'production';
const cwd = process.env.BRS_PROJECT_ROOT || process.cwd();

// Makes the script crash on unhandled rejections instead of silently
// ignoring them. In the future, promise rejections that are not handled will
// terminate the Node.js process with a non-zero exit code.
process.on('unhandledRejection', err => {
  throw err;
});

// Ensure environment variables are read.
require('../config/env');
const chalk = require('chalk');
const program = require('./utils/commander');
const Defer = require('./utils/defer');
// @remove-on-eject-begin
// Do the preflight checks (only happens before eject).
const verifyPackageTree = require('./utils/verifyPackageTree');
if (process.env.SKIP_PREFLIGHT_CHECK !== 'true') {
  verifyPackageTree();
}
// const verifyTypeScriptSetup = require('./utils/verifyTypeScriptSetup');
// verifyTypeScriptSetup();
// @remove-on-eject-end

const cp = require('child_process');
const path = require('path');
const fs = require('fs-extra');
const paths = require('../config/paths');
const checkRequiredFiles = require('react-dev-utils/checkRequiredFiles');
const printBuildError = require('react-dev-utils/printBuildError');
const FileSizeReporter = require('react-dev-utils/FileSizeReporter');

const measureFileSizesBeforeBuild =
  FileSizeReporter.measureFileSizesBeforeBuild;

const isInteractive = process.stdout.isTTY;

// Warn and crash if required files are missing
if (!checkRequiredFiles([paths.appHtml, paths.appIndexJs])) {
  process.exit(1);
}

// We require that you explicitly set browsers and do not fall back to
// browserslist defaults.
const { checkBrowsers } = require('react-dev-utils/browsersHelper');
checkBrowsers(paths.appPath, isInteractive)
  .then(() => {
    // First, read the current file sizes in build directory.
    // This lets us display how much they changed later.
    return measureFileSizesBeforeBuild(paths.appBuild);
  })
  .then(previousFileSizes => {
    // Remove all content but keep the directory so that
    // if you're in it, you don't end up in Trash

    if (program.ssr === true && !program.csr) {
      [path.resolve(program.out, 'static', 'bundle.ssr.js'), path.resolve(program.out, 'static', 'bundle.ssr.js.map')]
        .forEach(file => {
          try {
            fs.removeSync(file);
          } catch (e) {
            console.log(e);
          }
        });
    } else {
      fs.emptyDirSync(program.out);
    }

    // Merge with the public folder
    copyPublicFolder();
    // Start the webpack build
    return build(previousFileSizes);
  })
  .catch(err => {
    if (err && err.message) {
      console.log(err.message);
    }
    throw err;
  });
// Create the production build and print the deployment instructions.
function build(previousFileSizes) {
  // if has ssr.js or ssr.tsx, try to render a bundle.ssr.js for ssr.
  let ssrDefer = new Defer();
  let csrDefer = new Defer();
  let libraryDefer = new Defer();
  let libraryCP;
  let ssrCP;
  let csrCP;
  const tscCompileOnError = process.env.TSC_COMPILE_ON_ERROR === 'true';

  // We used to support resolving modules according to `NODE_PATH`.
  // This now has been deprecated in favor of jsconfig/tsconfig.json
  // This lets you use absolute paths in imports inside large monorepos:
  if (process.env.NODE_PATH) {
    console.log(
      chalk.yellow(
        'Setting NODE_PATH to resolve modules absolutely has been deprecated in favor of setting baseUrl in jsconfig.json (or tsconfig.json if you are using TypeScript) and will be removed in a future major release of create-react-app.'
      )
    );
    console.log();
  }

  if (!program.library && program.ssr) {
    console.log('Creating an optimized production build...');
    ssrCP = cp.fork(
      path.resolve(__dirname, './build-scripts/buildForSSR'),
      [
        program.out,
        program.no_hash === true ? '--no_hash' : '',
        program.modern ? '--modern' : '',
        program.snap ? '--snap' : ''
      ].filter(Boolean),
      {
        // execArgv: ['--inspect-brk=:9228'],
        cwd,
        env: process.env,
        stdio: 'inherit'
      }
    );
    ssrCP.on('error', e => {
      if (tscCompileOnError) {
        console.log(chalk.yellow(
          'SSR: Compiled with the following type errors (you may want to check these before deploying your app):\n'
        ));
        printBuildError(e);
      } else {
        console.log(chalk.red('SSR: Failed to compile.\n'));
        printBuildError(e);
        ssrDefer.reject(e);
      }
    });
    ssrCP.once('exit', code => {
      if (code === 0) {
        ssrDefer.resolve();
      } else {
        ssrDefer.reject();
      }
    });
  } else {
    ssrDefer.resolve();
  }

  if (!program.library && program.csr) {
    csrCP = cp.fork(
      path.resolve(__dirname, './build-scripts/buildForCSR'),
      [
        program.out,
        program.no_hash === true && '--no_hash',
        program.modern === true && '--modern',
        program.profile === true && '--profile'
      ].filter(Boolean),
      {
        cwd,
        env: {
          ...process.env,
          previousFileSizes: JSON.stringify(previousFileSizes)
        },
        stdio: 'inherit'
      }
    );
    csrCP.on('error', e => {
      if (tscCompileOnError) {
        console.log(chalk.yellow(
          'CSR: Compiled with the following type errors (you may want to check these before deploying your app):\n'
        ));
        printBuildError(e);
        csrDefer.resolve();
      } else {
        console.log(chalk.red('CSR: Failed to compile.\n'));
        printBuildError(e);
        csrDefer.reject(e);
      }
    });
    csrCP.once('exit', code => {
      if (code === 0) {
        csrDefer.resolve();
      } else {
        csrDefer.reject();
      }
    });
  } else {
    csrDefer.resolve();
  }

  if (program.library) {
    libraryCP = cp.fork(
      path.resolve(__dirname, './build-scripts/buildForLibrary'),
      [
        program.out,
        program.no_hash === true ? '--no_hash' : '',
        program.modern ? '--modern' : '',
        program.snap ? '--snap' : ''
      ].filter(Boolean),
      {
        cwd,
        env: {
          ...process.env,
          previousFileSizes: JSON.stringify(previousFileSizes)
        },
        stdio: 'inherit'
      }
    );
    libraryCP.on('error', e => {
      libraryDefer.reject(e);
    });
    libraryCP.once('exit', code => {
      if (code === 0) {
        libraryDefer.resolve();
      } else {
        libraryDefer.reject();
      }
    });
  } else {
    libraryDefer.resolve();
  }

  return Promise.all([
    csrDefer.promise,
    ssrDefer.promise,
    libraryDefer.promise
  ]).catch(e => {
    [ssrCP, csrCP, libraryCP].filter(Boolean).forEach(cp => {
      cp.kill(9);
    })
    throw e;
  });
}

function copyPublicFolder() {
  fs.copySync(paths.appPublic, paths.appBuild, {
    dereference: true,
    filter: file => file !== paths.appHtml
  });
}
