/* eslint-disable no-unused-vars */
'use strict';

const path = require('path');
const fs = require('fs-extra');
const ts = require('typescript');
const svgr = require('@svgr/core').default;
const SVGO = require('svgo');
const glob = require('glob');

const conflictMap = {
  es: {},
  lib: {},
};

async function converSvgFileToReactComponent(
  filepath,
  targetpath,
  useESModule
) {
  const parsedFilePath = path.parse(filepath);
  const componentName = parsedFilePath.name.replace(/(-\w)|(^\w)/g, v =>
    v.replace('-', '').toUpperCase()
  );
  const svgCode = await fs.readFile(filepath, 'utf8');
  let idPrefix = getSvgPrefix(parsedFilePath);
  const ns = useESModule ? 'es' : 'lib';
  if (conflictMap[ns][idPrefix]) {
    const i = ++conflictMap[ns][idPrefix];
    idPrefix = idPrefix + '-' + i;
  } else {
    conflictMap[ns][idPrefix] = 1;
  }

  const svgo = new SVGO({
    plugins: [{
      removeViewBox: false,
      prefixIds: {
        prefix: {
          toString() {
            return idPrefix;
          }
        }
      }
    }]
  });
  const { data } = await svgo.optimize(svgCode);
  let code = await svgr(
    data,
    {
      template: tsTemplate
    },
    { componentName }
  );

  // transform ts => js
  code = ts.transpile(code, {
    jsx: 'react',
    target: useESModule ? ts.ScriptTarget.ES2017 : ts.ScriptTarget.ES5,
    importHelpers: !useESModule,
    module: useESModule ? ts.ModuleKind.ES2015 : ts.ModuleKind.CommonJS
  });

  await fs.ensureDir(path.parse(targetpath).dir);
  return await fs.writeFile(targetpath, code, { encoding: 'utf8' });
}
exports.converSvgFileToReactComponent = converSvgFileToReactComponent;

async function convertSvgFiles(cwd, outdir, useESModule) {
  return new Promise((r, j) => {
    glob('**/*.svg', { cwd }, (_, matches) => {
      Promise.all(
        matches.map(file => {
          return converSvgFileToReactComponent(
            path.resolve(cwd, file),
            path.resolve(outdir, file + '.js'),
            useESModule
          );
        })
      ).then(r, j);
    });
  });
}

function tsTemplate({ template }, _opts, { componentName, jsx }) {
  const typeScriptTpl = template.smart({ plugins: ['typescript'] });
  return typeScriptTpl.ast`
    import React from 'react';
    export const ReactComponent = (props: React.SVGProps<SVGSVGElement>) => ${jsx};
    export default ReactComponent;
  `;
}

exports.default = convertSvgFiles;

function getSvgPrefix(parsedFilePath) {
  return parsedFilePath.dir.split(path.sep)
    .reverse()
    .filter((_, i) => i <= 2)
    .reverse()
    .join('_') + '_' + parsedFilePath.name;
}
